'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.run = run;
exports.caseOf = caseOf;
exports.assignArgs = assignArgs;

var _symbolObservable = require('symbol-observable');

var _symbolObservable2 = _interopRequireDefault(_symbolObservable);

var _Observable = require('rxjs/Observable');

var _Subject = require('rxjs/Subject');

require('rxjs/add/operator/startWith');

require('rxjs/add/operator/scan');

require('rxjs/add/operator/do');

require('rxjs/add/operator/distinctUntilChanged');

require('rxjs/add/operator/map');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function isEvent(e) {
  return e && e.preventDefault && e.stopPropagation;
}

function run(app) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var _app$main = app.main,
      init = _app$main.init,
      view = _app$main.view,
      update = _app$main.update;
  var onMsg = options.onMsg,
      onModel = options.onModel,
      onView = options.onView;

  var msg$ = new _Subject.Subject();
  var subscriptions = app.main.subscriptions;

  var initModel = init;
  var sideEffect = void 0;
  var cancelSubscriptions = void 0;

  if (typeof initModel === 'undefined') {
    throw new Error('Initial model is undefined. Pass null to explicitly set `no value`.');
  }

  if (Array.isArray(initModel) && typeof initModel[1] === 'function') {
    var _initModel = initModel;

    var _initModel2 = _slicedToArray(_initModel, 2);

    initModel = _initModel2[0];
    sideEffect = _initModel2[1];
  }

  var model$ = msg$.startWith(initModel).scan(function (model, msg) {
    if ((typeof msg === 'undefined' ? 'undefined' : _typeof(msg)) !== 'object') {
      throw new Error('Expected msg to be an object.');
    }

    onMsg && onMsg(msg);
    var newModel = update(msg, model);

    if (typeof newModel === 'undefined') {
      throw new Error('Unhandled msg ' + msg.constructor.name + ' ' + JSON.stringify(msg));
    }

    if (Array.isArray(newModel) && typeof newModel[1] === 'function') {
      var _newModel = newModel;

      var _newModel2 = _slicedToArray(_newModel, 2);

      newModel = _newModel2[0];
      sideEffect = _newModel2[1];
    }

    return newModel;
  })._do(function (model) {
    onModel && onModel(model);

    if (subscriptions) {
      cancelSubscriptions = subscriptions(dispatch, initModel);
      subscriptions = null;
    }

    if (sideEffect) {
      sideEffect(dispatch);
      sideEffect = null;
    }
  }).distinctUntilChanged();

  var view$ = view(dispatch, model$);
  if (typeof view$ === 'function') {
    view$ = model$.map(view$);
  }

  view$.subscribe(onView);

  function dispatch(msg) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    if (typeof msg === 'function') {
      return function eventHandler(e) {
        if (isEvent(e)) {
          var type = e.type,
              target = e.target;

          if (type === 'change' && target.tagName === 'INPUT' && (target.type === 'radio' || target.type === 'checkbox')) {
            msg$.next(new (Function.prototype.bind.apply(msg, [null].concat(args, [target.checked])))());
          } else {
            msg$.next(new (Function.prototype.bind.apply(msg, [null].concat(args, [target.value])))());
          }
        } else {
          msg$.next(new (Function.prototype.bind.apply(msg, [null].concat(Array.prototype.slice.call(arguments))))());
        }
      };
    }

    msg$.next(msg);
  }

  return {
    dispatch: dispatch,
    abort: function abort() {
      cancelSubscriptions && cancelSubscriptions();
      cancelSubscriptions = null;
      msg$.complete();
    }
  };
}

function caseOf(msg) {
  for (var i = 0; i < (arguments.length <= 1 ? 0 : arguments.length - 1); i += 2) {
    var Msg = arguments.length <= i + 1 ? undefined : arguments[i + 1];
    var update = arguments.length <= i + 1 + 1 ? undefined : arguments[i + 1 + 1];
    if (msg.constructor === Msg) {
      return update(msg);
    }
  }
}

var argNameCache = [];

function assignArgs(instance, args) {
  var argNames = argNameCache.find(function (c) {
    return c.msg === instance.constructor;
  });

  if (argNames) {
    argNames = argNames.argNames;
  } else {
    // https://davidwalsh.name/javascript-arguments
    argNames = instance.constructor.toString().match(/function\s.*?\(([^)]*)\)/)[1].split(',').map(function (arg) {
      return arg.replace(/\/\*.*\*\//, '').trim();
    }).filter(function (arg) {
      return arg;
    });

    argNameCache.push({ msg: instance.constructor, argNames: argNames });
  }

  var argMap = {};
  argNames.forEach(function (n, i) {
    return argMap[n] = args[i];
  });
  Object.assign(instance, argMap);
}